@@ -1,18 +1,20 @@
 # core
 import numpy as np
+import pandas as pd
 import random
 import json
 import itertools as it
 import logging
 import scipy.sparse
-from tqdm import tqdm
+#from tqdm import tqdm
 import time
-import importlib.resources as pkg_resources
+#import importlib.resources as pkg_resources
 
 # sklearn
 import sklearn
 from sklearn import *
 from sklearn.pipeline import Pipeline
+from sklearn.compose import ColumnTransformer
 
 # config space
 import ConfigSpace
@@ -24,7 +26,7 @@ from naiveautoml.commons import *
 
 class NaiveAutoML:
 
-    def __init__(self, search_space = None, scoring = None, num_cpus = 8, execution_timeout = 10, max_hpo_iterations = 100, timeout = None, standard_classifier=sklearn.neighbors.KNeighborsClassifier, logger_name = None, show_progress = False, opt_ordering = None, strictly_naive=False):
+    def __init__(self, search_space = None, scoring = None, side_scores = None , num_cpus = 8, execution_timeout = 10, max_hpo_iterations = 100, timeout = None, standard_classifier=sklearn.neighbors.KNeighborsClassifier, logger_name = None, show_progress = False, opt_ordering = None, strictly_naive=False, sparse = False):
         
         ''' search_space is a string or a list of dictionaries
             - if it is a dict, the last one for the learner and all the others for pre-processing. Each dictionary has an entry "name" and an entry "components", which is a list of components with their parameters.
@@ -48,6 +50,7 @@ class NaiveAutoML:
         else:
             self.search_space = search_space
         self.scoring = scoring
+        self.side_scores = side_scores
         self.num_cpus = num_cpus
         self.execution_timeout = execution_timeout
         self.max_hpo_iterations = max_hpo_iterations
@@ -60,6 +63,10 @@ class NaiveAutoML:
         self.stage_entrypoints = {}
         self.standard_classifier = standard_classifier
         
+        # mandatory pre-processing steps
+        self.sparse = sparse # do one-hot encoding via sparse representations (default is False since this is not supported by all algorithms)
+        self.mandatory_pre_processing = None
+        
         # define ordering of slots for optimization
         if opt_ordering is None:
             opt_ordering = ["classifier"]
@@ -69,11 +76,12 @@ class NaiveAutoML:
         self.opt_ordering = opt_ordering
         
         ## init logger
+        self.logger_name = logger_name
         self.logger = logging.getLogger('naiveautoml' if logger_name is None else logger_name)
         
     def check_combinations(self, X, y):
         
-        pool = EvaluationPool(X, y, self.scoring)
+        pool = EvaluationPool(X, y, self.scoring, self.side_scores)
         algorithms_per_stage = []
         names = []
         for step in self.search_space:
@@ -113,7 +121,7 @@ class NaiveAutoML:
                 steps = [("classifier", self.standard_classifier())]
             else:
                 steps = [(step_name, build_estimator(comp, None, X, y)), ("classifier", self.standard_classifier())]
-            return Pipeline(steps = steps)
+            return Pipeline(steps = self.mandatory_pre_processing + steps)
         
         else: ## semi-naive case (consider previous decisions)
             steps_tmp = [(s[0], build_estimator(s[1], None, X, y)) for s in decisions]
@@ -123,16 +131,18 @@ class NaiveAutoML:
             for step_inner in self.search_space:
                 if is_component_defined_in_steps(steps_tmp, step_inner["name"]):
                     steps_ordered.append(get_step_with_name(steps_tmp, step_inner["name"]))
-            return Pipeline(steps = steps_ordered)
+            return Pipeline(steps = self.mandatory_pre_processing + steps_ordered)
     
     
     def build_pipeline(self, hpo_processes, X, y):
         steps = self.get_instances_of_currently_selected_components_per_step(hpo_processes, X, y)
-        pl = Pipeline(steps)
+        pl = Pipeline(self.mandatory_pre_processing + steps)
         self.logger.debug(f"Original final pipeline is: {pl}")
+        i = 0
         while is_pipeline_forbidden(pl):
+            i += 1
             self.logger.debug("Invalid pipeline, removing first element!")
-            pl = Pipeline(steps=pl.steps[1:])
+            pl = Pipeline(steps=self.mandatory_pre_processing + steps[i:])
         return pl
     
     def choose_algorithms(self, X, y):
@@ -153,17 +163,21 @@ class NaiveAutoML:
             
             # create list of components to try for this slot
             step = [step for step in self.search_space if step["name"] == step_name][0]
-            self.logger.debug(f"Selecting component for step with name: {step_name}")
+            self.logger.info("--------------------------------------------------")
+            self.logger.info(f"Selecting component for step with name: {step_name}")
+            self.logger.info("--------------------------------------------------")
             if not step_name in ["classifier"]:
                 components = [None] + step["components"]
             else:
                 components = step["components"]
             
             # find best default parametrization for this slot (depending on choice of previously configured slots)
-            pool = EvaluationPool(X, y, self.scoring)
+            pool = EvaluationPool(X, y, self.scoring, self.side_scores, logger_name = self.logger_name + ".pool")
             best_score = -np.inf
             decision = None
             for comp in components:
+                if comp is not None and "MultinomialNB" in comp["class"]:
+                    continue
                 if self.deadline is not None:
                     remaining_time = self.deadline - 10 - time.time()
                     if remaining_time is not None and remaining_time < 0:
@@ -175,14 +189,15 @@ class NaiveAutoML:
                 # get and evaluate pipeline for this step
                 pl = self.get_pipeline_for_decision_in_step(step_name, comp, X, y, decisions)
                 try:
-                    score = pool.evaluate(pl, self.execution_timeout)
+                    scores = pool.evaluate(pl, min(self.execution_timeout, remaining_time if self.deadline is not None else 10**10))
                 except FunctionTimedOut:
                     self.logger.debug("TIMEOUT!")
-                    score = np.nan
+                    scores = {scoring: np.nan for scoring in [self.scoring] + (self.side_scores if self.side_scores is not None else [])}
+                score = scores[self.scoring]
                 self.logger.debug(f"Observed score of {score} for default configuration of {None if comp is None else comp['class']}")
                 
                 # update history
-                self.history.append({"time": time.time() - self.start_time, "pl": pl, "score_internal": score, "new_best": score > self.best_score_overall})
+                self.history.append({"time": time.time() - self.start_time, "pl": str(pl), "score_internal": score, "scores": scores, "new_best": score > self.best_score_overall})
                 
                 # update best score
                 if not np.isnan(score) and score > best_score:
@@ -240,7 +255,7 @@ class NaiveAutoML:
             else:
                 other_instances = [(step_name, None), ("classifier", self.standard_classifier())]
             index = 0 # it is (rather by coincidence) the first step we want to optimize
-            hpo = HPOProcess(step_name, comp, X, y, self.scoring, self.execution_timeout, other_instances, index, 1800, 1000)
+            hpo = HPOProcess(step_name, comp, X, y, self.scoring, self.side_scores, self.execution_timeout, self.mandatory_pre_processing, other_instances, index, 1800, 1000, allow_exhaustive_search = (self.max_hpo_iterations is None), logger_name = self.logger_name + ".hpo")
             hpo.best_score = components_with_score[step_name] # performance of default config
             hpo_processes[step_name] = hpo
         
@@ -269,13 +284,23 @@ class NaiveAutoML:
             for name in active_for_optimization:
                 hpo = hpo_processes[name]
                 self.logger.debug(f"Stepping HPO for {name}")
-                pl, score, runtime = hpo.step(remaining_time)
-                if score > self.best_score_overall:
-                    self.best_score_overall = score
-                self.history.append({"time": time.time() - self.start_time, "pl": pl, "score_internal": score, "new_best": score > self.best_score_overall})
-                if not hpo.active:
-                    self.logger.info(f"Deactivating {name}")
-                    inactive.append(name)
+                try:
+                    res = hpo.step(remaining_time)
+                    if res is not None:
+                        pl, scores, runtime = res
+                        score = scores[self.scoring]
+                        if score > self.best_score_overall:
+                            self.best_score_overall = score
+                        self.history.append({"time": time.time() - self.start_time, "pl": str(pl), "score_internal": score, "scores": scores, "new_best": score > self.best_score_overall})
+                        if not hpo.active:
+                            self.logger.info(f"Deactivating {name}")
+                            inactive.append(name)
+                except KeyboardInterrupt:
+                    raise
+                except Exception as e:
+                    self.logger.error(f"An error occurred in the HPO step: {e}")
+                    raise
+                    
             round_runtimes.append(time.time() - round_start)
             for name in inactive:
                 active_for_optimization.remove(name)
@@ -304,7 +329,30 @@ class NaiveAutoML:
         self.deadline = self.start_time + self.timeout if self.timeout is not None else None
         self.sparse_training_data = type(X) == scipy.sparse.csr.csr_matrix or type(X) == scipy.sparse.lil.lil_matrix
         if self.scoring is None:
-            self.scoring = "auc_roc" if len(np.unique(y)) == 2 else "neg_log_loss"
+            self.scoring = "roc_auc" if len(np.unique(y)) == 2 else "neg_log_loss"
+            
+        # determine fixed pre-processing steps for imputation and binarization
+        types = [set([type(v) for v in r]) for r in X.T]
+        numeric_features = [c for c, t in enumerate(types) if len(t) == 1 and list(t)[0] != str]
+        numeric_transformer = Pipeline([("imputer", sklearn.impute.SimpleImputer(strategy="median"))])
+        categorical_features = [i for i in range(X.shape[1]) if i not in numeric_features]
+        missing_values_per_feature = np.sum(pd.isnull(X), axis=0)
+        self.logger.info(f"There are {len(categorical_features)} categorical features, which will be binarized.")
+        self.logger.info(f"Missing values for the different attributes are {missing_values_per_feature}.")
+        if len(categorical_features) > 0 or sum(missing_values_per_feature) > 0:
+            categorical_transformer = Pipeline([
+                ("imputer", sklearn.impute.SimpleImputer(strategy="most_frequent")),
+                ("binarizer", sklearn.preprocessing.OneHotEncoder(handle_unknown='ignore', sparse = self.sparse)),
+
+            ])
+            self.mandatory_pre_processing = [("impute_and_binarize", ColumnTransformer(
+                transformers=[
+                    ("num", numeric_transformer, numeric_features),
+                    ("cat", categorical_transformer, categorical_features),
+                ]
+            ))]
+        else:
+            self.mandatory_pre_processing = []
         
         # print overview
         summary = ""
@@ -341,7 +389,7 @@ class NaiveAutoML:
         self.logger.info(f"Runtime was {self.end_time - self.start_time} seconds")
         
     def eval_history(self, X, y):
-        pool = EvaluationPool(X, y, self.scoring)
+        pool = EvaluationPool(X, y, self.scoring, self.side_scores)
         scores = []
         for entry in self.history:
             scores.append(pool.evaluate(entry["pl"]))
